From c04a7a95890ef5d032e6998675496bb438c3a14b Mon Sep 17 00:00:00 2001
From: Sebastian Kuzminsky <seb@lineratesystems.com>
Date: Mon, 22 Oct 2012 21:08:40 -0600
Subject: [PATCH 1/2] Update the mbuf csum_flags of IP fragments when
 computing their IP checksum

Before this commit, the ip_fragment() function does not clear the
mbuf CSUM_IP flag ("this mbuf needs an IP header checksum"), even
when it computes the IP header checksum itself.  This behavior is
acceptable when ip_fragment() is called from ip_output(), because
ip_output() clears the mbuf's flag.  But it is not acceptable when
ip_fragment() is called from ip_fastforward(), because ip_fastforward()
does not clear the mbuf's flag.

The result is that, when forwarding a packet that needs fragmentation,
and the fragments are sent by a NIC that does not advertise hardware
IP checksum offloading, and that NIC *does* check for for the CSUM_IP
flag anyway and then gets the IP checksum wrong, *then* the fragments
going out on the wire would have the wrong checksum.

The em driver does not advertise IP header checksum offloading, but
does try to set up IP header checksum offloading anyway when the
mbuf is marked CSUM_IP, and it gets the IP checksum wrong.

The fix is to clear the CSUM_IP flag in the mbuf in ip_fragment()
when the IP checksum is computed, to let the lower layers know that
they don't need to do it.
---
 sys/netinet/ip_fastfwd.c |    9 ++++++++-
 sys/netinet/ip_output.c  |   30 ++++++++++++++++++++++++++++--
 2 files changed, 36 insertions(+), 3 deletions(-)

diff --git a/sys/netinet/ip_fastfwd.c b/sys/netinet/ip_fastfwd.c
index e84695e..ae65bfe 100644
--- a/sys/netinet/ip_fastfwd.c
+++ b/sys/netinet/ip_fastfwd.c
@@ -558,12 +558,19 @@ passout:
 			goto consumed;
 		} else {
 			/*
-			 * We have to fragment the packet
+			 * We have to fragment this packet, and the fragments
+			 * will need all-new IP checksums.  (The payload
+			 * checksums, if any, don't need to be modified
+			 * because the payload will be reassembled before
+			 * delivery.)
 			 */
 			m->m_pkthdr.csum_flags |= CSUM_IP;
 			/*
 			 * ip_fragment expects ip_len and ip_off in host byte
 			 * order but returns all packets in network byte order
+			 * If if_hwassist doesn't advertise IP checksum
+			 * offloading, ask ip_fragment to do it for us in
+			 * software now.
 			 */
 			if (ip_fragment(ip, &m, mtu, ifp->if_hwassist,
 					(~ifp->if_hwassist & CSUM_DELAY_IP))) {
diff --git a/sys/netinet/ip_output.c b/sys/netinet/ip_output.c
index adbb074..08c4185 100644
--- a/sys/netinet/ip_output.c
+++ b/sys/netinet/ip_output.c
@@ -506,12 +506,30 @@ passout:
 		}
 	}
 
+	/* Annotate the outgoing packet: it needs its IP header checksummed. */
 	m->m_pkthdr.csum_flags |= CSUM_IP;
+
+	/* sw_csum is everything the packet needs that *won't* be done in
+	 * hardware.
+	 */
 	sw_csum = m->m_pkthdr.csum_flags & ~ifp->if_hwassist;
+
+	/* Do payload checksumming in software, now, if needed & wanted. */
 	if (sw_csum & CSUM_DELAY_DATA) {
 		in_delayed_cksum(m);
 		sw_csum &= ~CSUM_DELAY_DATA;
 	}
+
+	/* Clear all the packet's needs that'll be done by software.
+	 * At this point the packet's needs are (m_pkthdr.csum_flags | sw_csum),
+	 * and software should do the stuff in sw_csum.
+	 *
+	 * FIXME: This is a bug, stuff in the code paths after this
+	 * (for example ip_fragment) expect m_pkthdr->csum_flags to be the
+	 * list of stuff the packet needs.  in_delayed_cksum() above also
+	 * has this expectation, which is why this code is convoluted to
+	 * call it before clearing m's csum_flags.
+	 */
 	m->m_pkthdr.csum_flags &= ifp->if_hwassist;
 
 	/*
@@ -526,6 +544,10 @@ passout:
 		ip->ip_sum = 0;
 		if (sw_csum & CSUM_DELAY_IP)
 			ip->ip_sum = in_cksum(m, hlen);
+			/* Normally we'd clear CSUM_DELAY_IP out of sw_csum
+			 * here, but that variable is not used again before
+			 * it passes out of scope.
+			 */
 
 		/*
 		 * Record statistics for this interface address.
@@ -743,8 +765,10 @@ smart_frag_failure:
 		m->m_pkthdr.csum_flags = m0->m_pkthdr.csum_flags;
 		mhip->ip_off = htons(mhip->ip_off);
 		mhip->ip_sum = 0;
-		if (sw_csum & CSUM_DELAY_IP)
+		if (sw_csum & CSUM_DELAY_IP) {
 			mhip->ip_sum = in_cksum(m, mhlen);
+			m->m_pkthdr.csum_flags &= ~CSUM_DELAY_IP;
+		}
 		*mnext = m;
 		mnext = &m->m_nextpkt;
 	}
@@ -764,8 +788,10 @@ smart_frag_failure:
 	ip->ip_off |= IP_MF;
 	ip->ip_off = htons(ip->ip_off);
 	ip->ip_sum = 0;
-	if (sw_csum & CSUM_DELAY_IP)
+	if (sw_csum & CSUM_DELAY_IP) {
 		ip->ip_sum = in_cksum(m0, hlen);
+		m0->m_pkthdr.csum_flags &= ~CSUM_DELAY_IP;
+	}
 
 done:
 	*m_frag = m0;
-- 
1.7.8.3

